import numpy as np
from pathlib import Path
from typing import Optional, List, Dict, Union

import autoarray as aa

from autocti.charge_injection.imaging.settings import SettingsImagingCI
from autocti.charge_injection.layout import Layout2DCI
from autocti.extract.settings import SettingsExtract
from autocti.mask import mask_2d
from autocti import exc


class ImagingCI(aa.Imaging):
    def __init__(
        self,
        data: aa.Array2D,
        noise_map: aa.Array2D,
        pre_cti_data: aa.Array2D,
        layout: Layout2DCI,
        cosmic_ray_map: Optional[aa.Array2D] = None,
        noise_scaling_map_dict: Optional[Dict] = None,
        fpr_value: Optional[float] = None,
        settings_dict: Optional[Dict] = None,
    ):
        super().__init__(data=data, noise_map=noise_map)

        self.data = self.data.native
        self.noise_map = self.noise_map.native
        self.pre_cti_data = pre_cti_data.native

        if cosmic_ray_map is not None:
            cosmic_ray_map = cosmic_ray_map.native

        self.cosmic_ray_map = cosmic_ray_map

        if noise_scaling_map_dict is not None:
            noise_scaling_map_dict = {
                key: noise_scaling_map.native
                for key, noise_scaling_map in noise_scaling_map_dict.items()
            }

        self.noise_scaling_map_dict = noise_scaling_map_dict

        self.layout = layout

        if fpr_value is None:
            fpr_value = np.round(
                np.mean(
                    self.layout.extract.parallel_fpr.median_list_from(
                        array=self.data,
                        settings=SettingsExtract(
                            pixels_from_end=min(
                                10, self.layout.smallest_parallel_rows_within_ci_regions
                            )
                        ),
                    )
                ),
                2,
            )

        self.fpr_value = fpr_value
        self.settings_dict = settings_dict

    @property
    def mask(self):
        return self.data.mask

    @property
    def region_list(self):
        return self.layout.region_list

    @property
    def norm_columns_list(self) -> List:
        """
        The `layout` describes the 2D regions on the data containing charge whose input signal properties are know
        beforehand (e.g. charge injection imaging).

        However, the exact values may not be known and therefore need to be estimated from the image.

        This function estimates the normalization of every column of data in the 2D regions, by taking the median
        of each column. If a mask is applied (e.g. to remove cosmic rays) these pixels are omitted from the median.

        Returns
        -------
        A list of the normalization of every column of the charge regions
        """
        masked_image = np.ma.array(data=self.data, mask=self.data.mask)

        return [
            np.ma.median(masked_image[region.y0 : region.y1, column_index])
            for region in self.region_list
            for column_index in range(region.x0, region.x1)
        ]

    @property
    def pre_cti_data_residual_map(self) -> aa.Array2D:
        """
        The residuals of the data and the pre CTI data.

        This is used to assess whether the pre CTI data has been estimated accurately (e.g. from the FPR of the
        data) and includes e specific set of visualization functions.

        Returns
        -------
        The residual map of the data and pre CTI data.
        """
        return self.data - self.pre_cti_data

    def apply_mask(self, mask: mask_2d.Mask2D) -> "ImagingCI":
        image = aa.Array2D(values=self.data.native, mask=mask)
        noise_map = aa.Array2D(values=self.noise_map.native, mask=mask)

        if self.cosmic_ray_map is not None:
            cosmic_ray_map = aa.Array2D(values=self.cosmic_ray_map.native, mask=mask)

        else:
            cosmic_ray_map = None

        if self.noise_scaling_map_dict is not None:
            noise_scaling_map_dict = {
                key: aa.Array2D(values=noise_scaling_map.native, mask=mask)
                for key, noise_scaling_map in self.noise_scaling_map_dict.items()
            }

        else:
            noise_scaling_map_dict = None

        return ImagingCI(
            data=image,
            noise_map=noise_map,
            pre_cti_data=self.pre_cti_data.native,
            layout=self.layout,
            cosmic_ray_map=cosmic_ray_map,
            noise_scaling_map_dict=noise_scaling_map_dict,
            fpr_value=self.fpr_value,
            settings_dict=self.settings_dict,
        )

    def apply_settings(self, settings: SettingsImagingCI):
        if settings.parallel_pixels is not None:
            dataset = self.layout.extract.parallel_calibration.imaging_ci_from(
                dataset=self, columns=settings.parallel_pixels
            )

            mask = self.layout.extract.parallel_calibration.mask_2d_from(
                mask=self.mask, columns=settings.parallel_pixels
            )

        elif settings.serial_pixels is not None:
            dataset = self.layout.extract.serial_calibration.imaging_ci_from(
                dataset=self, rows=settings.serial_pixels
            )

            mask = self.layout.extract.serial_calibration.mask_2d_from(
                mask=self.mask, rows=settings.serial_pixels
            )

        else:
            return self

        dataset = dataset.apply_mask(mask=mask)

        return dataset

    def set_noise_scaling_map_dict(self, noise_scaling_map_dict: Dict):
        self.noise_scaling_map_dict = {
            key: noise_scaling_map.native
            for key, noise_scaling_map in noise_scaling_map_dict.items()
        }

    @classmethod
    def from_fits(
        cls,
        pixel_scales: aa.type.PixelScales,
        layout: Layout2DCI,
        data_path: Optional[Union[Path, str]] = None,
        data_hdu: int = 0,
        data: aa.Array2D = None,
        noise_map_path: Optional[Union[Path, str]] = None,
        noise_map_hdu: int = 0,
        noise_map_from_single_value: float = None,
        pre_cti_data_path: Optional[Union[Path, str]] = None,
        pre_cti_data_hdu: int = 0,
        pre_cti_data: aa.Array2D = None,
        cosmic_ray_map_path: Optional[Union[Path, str]] = None,
        cosmic_ray_map_hdu: int = 0,
        settings_dict: Optional[Dict] = None,
    ) -> "ImagingCI":
        """
        Load charge injection imaging from multiple .fits file.

        For each attribute of the charge injection data (e.g. `data`, `noise_map`, `pre_cti_data`) the path to
        the .fits and the `hdu` containing the data can be specified.

        The `noise_map` assumes the noise value in each `data` value are independent, where these values are the
        RMS standard deviation error in each pixel.

        If the dataset has a mask associated with it (e.g. in a `mask.fits` file) the file must be loaded separately
        via the `Mask2D` object and applied to the imaging after loading via fits using the `from_fits` method.

        Parameters
        ----------
        pixel_scales
            The (y,x) arcsecond-to-pixel units conversion factor of every pixel. If this is input as a `float`,
            it is converted to a (float, float).
        layout
            The layout of the charge injection, containing information like where the parallel and serial FPR and
            EPER are located.
        data_path
            The path to the data .fits file containing the image data (e.g. '/path/to/data.fits').
        data_hdu
            The hdu the image data is contained in the .fits file specified by `data_path`.
        data
            Manually input the data as an `Array2D` instead of loading it via a .fits file.
        noise_map_path
            The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits').
        noise_map_hdu
            The hdu the noise map is contained in the .fits file specified by `noise_map_path`.
        noise_map_from_single_value
            Creates a `noise_map` of constant values if this is input instead of loading via .fits.
        pre_cti_data_path
            The path to the pre CTI data .fits file containing the image data (e.g. '/path/to/pre_cti_data.fits').
        pre_cti_data_hdu
            The hdu the pre cti data is contained in the .fits file specified by `pre_cti_data_path`.
        pre_cti_data
            Manually input the pre CTI data as an `Array2D` instead of loading it via a .fits file.
        cosmic_ray_map_path
            The path to the cosmic ray map .fits file containing the map of cosmic
            rays (e.g. '/path/to/cosmic_ray_map.fits').
        cosmic_ray_map_hdu
            The hdu the cosmic ray data is contained in the .fits file specified by `cosmic_ray_map_path`.
        settings_dict
            A dictionary of settings associated with the charge injeciton imaging (e.g. voltage settings) which is
            used for visualization.
        """
        if data_path is not None and data is None:
            data = aa.Array2D.from_fits(
                file_path=data_path, hdu=data_hdu, pixel_scales=pixel_scales
            )

        if noise_map_path is not None:
            noise_map = aa.util.array_2d.numpy_array_2d_via_fits_from(
                file_path=noise_map_path, hdu=noise_map_hdu
            )
        else:
            noise_map = np.ones(data.shape_native) * noise_map_from_single_value

        noise_map = aa.Array2D.no_mask(values=noise_map, pixel_scales=pixel_scales)

        if pre_cti_data_path is not None and pre_cti_data is None:
            pre_cti_data = aa.Array2D.from_fits(
                file_path=pre_cti_data_path,
                hdu=pre_cti_data_hdu,
                pixel_scales=pixel_scales,
            )
        elif pre_cti_data is None:
            raise exc.ImagingCIException(
                "Cannot load pre_cti_data from .fits and pass explicit pre_cti_data."
            )

        pre_cti_data = aa.Array2D.no_mask(
            values=pre_cti_data.native, pixel_scales=pixel_scales
        )

        if cosmic_ray_map_path is not None:
            cosmic_ray_map = aa.Array2D.from_fits(
                file_path=cosmic_ray_map_path,
                hdu=cosmic_ray_map_hdu,
                pixel_scales=pixel_scales,
            )

        else:
            cosmic_ray_map = None

        return ImagingCI(
            data=data,
            noise_map=noise_map,
            pre_cti_data=pre_cti_data,
            cosmic_ray_map=cosmic_ray_map,
            layout=layout,
            settings_dict=settings_dict,
        )

    def output_to_fits(
        self,
        data_path: Union[Path, str],
        noise_map_path: Optional[Union[Path, str]] = None,
        pre_cti_data_path: Optional[Union[Path, str]] = None,
        cosmic_ray_map_path: Optional[Union[Path, str]] = None,
        overwrite: bool = False,
    ):
        """
        Output the charge injection imaging dataset to multiple .fits file.

        For each attribute of the charge injection imaging data (e.g. `data`, `noise_map`, `pre_cti_data`) the path to
        the .fits can be specified, with `hdu=0` assumed automatically.

        If the `data` has been masked, the masked data is output to .fits files. A mask can be separately output to
        a file `mask.fits` via the `Mask` objects `output_to_fits` method.

        Parameters
        ----------
        data_path
            The path to the data .fits file where the image data is output (e.g. '/path/to/data.fits').
        noise_map_path
            The path to the noise_map .fits where the noise_map is output (e.g. '/path/to/noise_map.fits').
        pre_cti_data_path
            The path to the pre CTI data .fits file where the pre CTI data is output (e.g. '/path/to/pre_cti_data.fits').
        cosmic_ray_map_path
            The path to the cosmic ray map .fits file where the cosmic ray map is
            output (e.g. '/path/to/cosmic_ray_map.fits').
        overwrite
            If `True`, the .fits files are overwritten if they already exist, if `False` they are not and an
            exception is raised.
        """
        self.data.output_to_fits(file_path=data_path, overwrite=overwrite)

        if noise_map_path is not None:
            self.noise_map.output_to_fits(file_path=noise_map_path, overwrite=overwrite)

        if pre_cti_data_path is not None:
            self.pre_cti_data.output_to_fits(
                file_path=pre_cti_data_path, overwrite=overwrite
            )

        if self.cosmic_ray_map is not None and cosmic_ray_map_path is not None:
            self.cosmic_ray_map.output_to_fits(
                file_path=cosmic_ray_map_path, overwrite=overwrite
            )



from typing import Optional, List

from autoconf.dictable import from_json

from arcticpy import CCDPhase
from arcticpy import TrapInstantCapture


class AbstractCTI:
    @staticmethod
    def from_json(file_path):
        return from_json(file_path=file_path)

    @property
    def trap_list(self):
        raise NotImplementedError

    @property
    def delta_ellipticity(self):
        return sum([trap.delta_ellipticity for trap in self.trap_list])


class CTI1D(AbstractCTI):
    def __init__(
        self,
        trap_list: Optional[List[TrapInstantCapture]] = None,
        ccd: Optional[CCDPhase] = None,
    ):
        """
        An object which determines the behaviour of CTI during 1D clocking.

        This includes the traps that capture and trail electrons and the CCD volume filling behaviour.

        Parameters
        ----------
        trap_list
            The traps on the dataset that capture and release electrons during clocking.
        ccd
            The CCD volume filling parameterization which dictates how an electron cloud fills pixels and thus
            how it is subject to traps.
        """
        self._trap_list = trap_list
        self.ccd = ccd

    @property
    def trap_list(self):
        return self._trap_list


class CTI2D(AbstractCTI):
    def __init__(
        self,
        parallel_trap_list: Optional[List[TrapInstantCapture]] = None,
        parallel_ccd: Optional[CCDPhase] = None,
        serial_trap_list: Optional[List[TrapInstantCapture]] = None,
        serial_ccd: Optional[CCDPhase] = None,
    ):
        """
        An object which determines the behaviour of CTI during 2D parallel and serial clocking.

        This includes the traps that capture and trail electrons and the CCD volume filling behaviour.

        Parameters
        ----------
        parallel_trap_list
            The traps on the dataset that capture and release electrons during parallel clocking.
        parallel_ccd
            The CCD volume filling parameterization which dictates how an electron cloud fills pixel in the parallel
             direction and thus how it is subject to traps.
        serial_trap_list
            The traps on the dataset that capture and release electrons during serial clocking.
        serial_ccd
            The CCD volume filling parameterization which dictates how an electron cloud fills pixel in the serial
             direction and thus how it is subject to traps.
        """
        self.parallel_trap_list = parallel_trap_list
        self.parallel_ccd = parallel_ccd
        self.serial_trap_list = serial_trap_list
        self.serial_ccd = serial_ccd

    @property
    def trap_list(self) -> List[TrapInstantCapture]:
        """
        Combine the parallel and serial trap lists to make an overall list of traps in the model.

        This is not a straight forward list addition, because **PyAutoFit** model's store the `parallel_traps` and
        `serial_traps` entries as a `ModelInstance`. This object does not allow for straight forward list addition.
        """
        parallel_traps = self.parallel_trap_list or []
        serial_traps = self.serial_trap_list or []

        return [trap for trap in parallel_traps] + [trap for trap in serial_traps]


def is_parallel_fit(model):
    return model.parallel_ccd is not None and model.serial_ccd is None


def is_serial_fit(model):
    return model.parallel_ccd is None and model.serial_ccd is not None


def is_parallel_and_serial_fit(model):
    return model.parallel_ccd is not None and model.serial_ccd is not None



import numpy as np
from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.one_d.abstract import Extract1D
from autocti.extract.settings import SettingsExtract


class Extract1DEPER(Extract1D):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region1D]:
        """
        Returns a list of the (x0, x1) regions containing the EPERs of a 1D CTI dataset.

        These are used for extracting the EPER regions of 1D data.

        Negative pixel values can be input into the `pixels` tuple, whereby pixels in front of the EPERs (e.g.
        the FPR) are extracted.

        Parameters
        ----------
        pixels
            The row indexes to extract the trails between (e.g. pixels(0, 3) extracts the 1st, 2nd and 3rd pixels)
        """

        pixels = settings.pixels

        region_list = []

        for i, region in enumerate(self.region_list):
            if settings.pixels_from_end is not None:
                parallel_row_spaces = self.parallel_rows_between_regions + [
                    self.trail_size_to_array_edge
                ]

                pixels = (
                    parallel_row_spaces[i] - settings.pixels_from_end,
                    parallel_row_spaces[i],
                )

            region_list.append(region.trailing_region_from(pixels=pixels))

        return region_list

    def array_1d_from(self, array: aa.Array1D) -> aa.Array1D:
        """
        Extract all of the data values in an input `array1D` that do not overlap the charge regions or the
        prescan / overscan regions.

        This  extracts a `array1D` that contains only regions of the data where there are parallel trails (e.g. those
        that follow the charge-injection regions).
        """

        array_1d_eper = array.native.copy()

        for region in self.region_list:
            array_1d_eper[region.slice] = 0.0

        array_1d_eper.native[self.prescan.slice] = 0.0
        array_1d_eper.native[self.overscan.slice] = 0.0

        return array_1d_eper



from typing import Optional

from arcticpy import add_cti
from arcticpy import remove_cti

from arcticpy import ROE

import autoarray as aa

from autocti.clocker.abstract import AbstractClocker
from autocti.model.model_util import CTI1D


class Clocker1D(AbstractClocker):
    def __init__(
        self,
        iterations: int = 5,
        roe: Optional[ROE] = None,
        express: int = 0,
        window_start: int = 0,
        window_stop: int = -1,
        time_start=0,
        time_stop=-1,
        prune_n_electrons=1e-18,
        prune_frequency=20,
        allow_negative_pixels=1,
        pixel_bounce=None,
        verbosity: int = 0,
    ):
        """
        Performs clocking of a 1D signal via the c++ arctic algorithm.

        This corresponds to a single row or column of a CCD in the parallel or serial direction. Given the notion of
        parallel and serial are not relevent in 1D, these prefixes are dropped from parameters (unlike the `Clocker2D`)
        object.

        Parameters
        ----------
        iterations
            The number of iterations used to correct CTI from an image.
        roe
            Contains parameters describing the read-out electronics of the CCD (e.g. CCD dwell times, charge injection
            clocking, etc.).
        express
            An integer factor describing how pixel-to-pixel transfers are combined into single transfers for
            efficiency (see: https://academic.oup.com/mnras/article/401/1/371/1006825).
        window_start
            The pixel index of the input image where arCTIc clocking begins, for example if `window_start=10` the
            first 10 pixels are omitted and not clocked.
        window_start
            The pixel index of the input image where arCTIc clocking ends, for example if `window_start=20` any
            pixels after the 20th pixel are omitted and not clocked.
        verbosity
            Whether to silence print statements and output from the c++ arctic call.
        """

        super().__init__(iterations=iterations, verbosity=verbosity)

        self.roe = roe or ROE()
        self.express = express
        self.window_start = window_start
        self.window_stop = window_stop
        self.time_start = time_start
        self.time_stop = time_stop
        self.prune_n_electrons = prune_n_electrons
        self.prune_frequency = prune_frequency

        self.allow_negative_pixels = allow_negative_pixels
        self.pixel_bounce = pixel_bounce

    def _traps_ccd_from(self, cti: CTI1D):
        """
        Unpack the `CTI1D` object to retries its traps and ccd.
        """
        if cti.trap_list is not None:
            trap_list = list(cti.trap_list)
        else:
            trap_list = None

        return trap_list, cti.ccd

    def add_cti(self, data: aa.Array1D, cti: CTI1D) -> aa.Array1D:
        """
        Add CTI to a 1D dataset by passing it from the c++ arctic clocking algorithm.

        The c++ arctic wrapper takes as input a 2D ndarray, therefore this function converts the input 1D data to
        a 2D ndarray where the second dimension is 1, passes this from arctic and then flattens the 2D ndarray that
        is returned by to a 1D `Array1D` object with the input data's dimensions.

        Parameters
        ----------
        data
            The 1D data that is clocked via arctic and has CTI added to it.
        cti
            An object which represents the CTI properties of 1D clocking, including the trap species which capture
            and release electrons and the volume-filling behaviour of the CCD.
        """

        trap_list, ccd = self._traps_ccd_from(cti=cti)

        self.check_traps(trap_list_0=trap_list)
        self.check_ccd(ccd_list=[ccd])

        image_pre_cti_2d = aa.Array2D.zeros(
            shape_native=(data.shape_native[0], 1), pixel_scales=data.pixel_scales
        ).native

        image_pre_cti_2d[:, 0] = data

        ccd = self.ccd_from(ccd_phase=ccd)

        try:
            image_post_cti = add_cti(
                image=image_pre_cti_2d,
                parallel_ccd=ccd,
                parallel_roe=self.roe,
                parallel_traps=trap_list,
                parallel_express=self.express,
                parallel_window_offset=data.readout_offsets[0],
                parallel_window_start=self.window_start,
                parallel_window_stop=self.window_stop,
                parallel_time_start=self.time_start,
                parallel_time_stop=self.time_stop,
                parallel_prune_n_electrons=self.prune_n_electrons,
                parallel_prune_frequency=self.prune_frequency,
                allow_negative_pixels=self.allow_negative_pixels,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )
        except TypeError:
            image_post_cti = add_cti(
                image=image_pre_cti_2d,
                parallel_ccd=ccd,
                parallel_roe=self.roe,
                parallel_traps=trap_list,
                parallel_express=self.express,
                parallel_window_offset=data.readout_offsets[0],
                parallel_window_start=self.window_start,
                parallel_window_stop=self.window_stop,
                parallel_time_start=self.time_start,
                parallel_time_stop=self.time_stop,
                parallel_prune_n_electrons=self.prune_n_electrons,
                parallel_prune_frequency=self.prune_frequency,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )

        return aa.Array1D.no_mask(
            values=image_post_cti.flatten(), pixel_scales=data.pixel_scales
        )

    def remove_cti(self, data: aa.Array1D, cti: CTI1D) -> aa.Array1D:
        """
        Add CTI to a 1D dataset by passing it from the c++ arctic clocking algorithm.

        The c++ arctic wrapper takes as input a 2D ndarray, therefore this function converts the input 1D data to
        a 2D ndarray where the second dimension is 1, passes this from arctic and then flattens the 2D ndarray that
        is returned by to a 1D `Array1D` object with the input data's dimensions.

        Parameters
        ----------
        data
            The 1D data that is clocked via arctic and has CTI removeed to it.
        cti
            An object which represents the CTI properties of the clocking, including the trap species which capture
            and release electrons and the volume-filling behaviour of the CCD.
        """

        trap_list, ccd = self._traps_ccd_from(cti=cti)

        self.check_traps(trap_list_0=trap_list)
        self.check_ccd(ccd_list=[ccd])

        image_pre_cti_2d = aa.Array2D.zeros(
            shape_native=(data.shape_native[0], 1), pixel_scales=data.pixel_scales
        ).native

        image_pre_cti_2d[:, 0] = data

        ccd = self.ccd_from(ccd_phase=ccd)

        try:
            image_post_cti = remove_cti(
                image=image_pre_cti_2d,
                n_iterations=self.iterations,
                parallel_ccd=ccd,
                parallel_roe=self.roe,
                parallel_traps=trap_list,
                parallel_express=self.express,
                parallel_window_offset=data.readout_offsets[0],
                parallel_window_start=self.window_start,
                parallel_window_stop=self.window_stop,
                parallel_time_start=self.time_start,
                parallel_time_stop=self.time_stop,
                parallel_prune_n_electrons=self.prune_n_electrons,
                parallel_prune_frequency=self.prune_frequency,
                allow_negative_pixels=self.allow_negative_pixels,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )
        except TypeError:
            image_post_cti = remove_cti(
                image=image_pre_cti_2d,
                n_iterations=self.iterations,
                parallel_ccd=ccd,
                parallel_roe=self.roe,
                parallel_traps=trap_list,
                parallel_express=self.express,
                parallel_window_offset=data.readout_offsets[0],
                parallel_window_start=self.window_start,
                parallel_window_stop=self.window_stop,
                parallel_time_start=self.time_start,
                parallel_time_stop=self.time_stop,
                parallel_prune_n_electrons=self.prune_n_electrons,
                parallel_prune_frequency=self.prune_frequency,
                pixel_bounce=self.pixel_bounce,
                verbosity=self.verbosity,
            )

        return aa.Array1D.no_mask(
            values=image_post_cti.flatten(), pixel_scales=data.pixel_scales
        )



import numpy as np
from pathlib import Path
from typing import Optional, Dict, Union

import autoarray as aa

from autocti import exc
from autocti.dataset_1d.dataset_1d.settings import SettingsDataset1D
from autocti.extract.settings import SettingsExtract
from autocti.layout.one_d import Layout1D


class Dataset1D(aa.AbstractDataset):
    def __init__(
        self,
        data: aa.Array1D,
        noise_map: aa.Array1D,
        pre_cti_data: aa.Array1D,
        layout: Layout1D,
        fpr_value: Optional[float] = None,
        settings: SettingsDataset1D = SettingsDataset1D(),
        settings_dict: Optional[Dict] = None,
    ):
        super().__init__(data=data, noise_map=noise_map, settings=settings)

        self.data = data
        self.noise_map = noise_map
        self.pre_cti_data = pre_cti_data
        self.layout = layout

        if fpr_value is None:
            fpr_value = np.round(
                np.median(
                    self.layout.extract.fpr.stacked_array_1d_from(
                        array=self.data,
                        settings=SettingsExtract(
                            pixels_from_end=min(
                                5, self.layout.extract.fpr.total_pixels_min
                            )
                        ),
                    )
                ),
                2,
            )

        self.fpr_value = fpr_value

        self.settings_dict = settings_dict

    def apply_mask(self, mask: aa.Mask1D) -> "Dataset1D":
        data = aa.Array1D(values=self.data, mask=mask).native
        noise_map = aa.Array1D(values=self.noise_map.astype("float"), mask=mask).native

        return Dataset1D(
            data=data,
            noise_map=noise_map,
            pre_cti_data=self.pre_cti_data,
            layout=self.layout,
            fpr_value=self.fpr_value,
            settings_dict=self.settings_dict,
        )

    def apply_settings(self, settings: SettingsDataset1D) -> "Dataset1D":
        return self

    @classmethod
    def from_fits(
        cls,
        pixel_scales: aa.type.PixelScales,
        layout: Layout1D,
        data_path: Optional[Union[Path, str]] = None,
        data_hdu: int = 0,
        noise_map_path: Optional[Union[Path, str]] = None,
        noise_map_hdu: int = 0,
        noise_map_from_single_value: float = None,
        pre_cti_data_path: Optional[Union[Path, str]] = None,
        pre_cti_data_hdu: int = 0,
        pre_cti_data: aa.Array1D = None,
        settings_dict: Optional[Dict] = None,
    ):
        """
        Load 1D dataset from multiple .fits file.

        For each attribute of the 1D dataset (e.g. `data`, `noise_map`, `pre_cti_data`) the path to the .fits and
        the `hdu` containing the data can be specified.

        The `noise_map` assumes the noise value in each `data` value are independent, where these values are the
        RMS standard deviation error in each pixel.

        If the dataset has a mask associated with it (e.g. in a `mask.fits` file) the file must be loaded separately
        via the `Mask1D` object and applied to the imaging after loading via fits using the `from_fits` method.

        Parameters
        ----------
        pixel_scales
            The (y,x) arcsecond-to-pixel units conversion factor of every pixel. If this is input as a `float`,
            it is converted to a (float, float).
        layout
            The layout of the 1D dataset, containing information like where the FPR and EPER are located.
        data_path
            The path to the data .fits file containing the data (e.g. '/path/to/data.fits').
        data_hdu
            The hdu the image data is contained in the .fits file specified by `data_path`.
        noise_map_path
            The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits').
        noise_map_hdu
            The hdu the noise map is contained in the .fits file specified by `noise_map_path`.
        noise_map_from_single_value
            Creates a `noise_map` of constant values if this is input instead of loading via .fits.
        pre_cti_data_path
            The path to the pre CTI data .fits file containing the image data (e.g. '/path/to/pre_cti_data.fits').
        pre_cti_data_hdu
            The hdu the pre cti data is contained in the .fits file specified by `pre_cti_data_path`.
        pre_cti_data
            Manually input the pre CTI data as an `Array1D` instead of loading it via a .fits file.
        settings_dict
            A dictionary of settings associated with the charge injeciton imaging (e.g. voltage settings) which is
            used for visualization.
        """
        data = aa.Array1D.from_fits(
            file_path=data_path, hdu=data_hdu, pixel_scales=pixel_scales
        )

        if noise_map_path is not None:
            noise_map = aa.util.array_1d.numpy_array_1d_via_fits_from(
                file_path=noise_map_path, hdu=noise_map_hdu
            ).astype("float")
        else:
            noise_map = np.ones(data.shape_native) * noise_map_from_single_value

        noise_map = aa.Array1D.no_mask(values=noise_map, pixel_scales=pixel_scales)

        if pre_cti_data_path is not None and pre_cti_data is None:
            pre_cti_data = aa.Array1D.from_fits(
                file_path=pre_cti_data_path,
                hdu=pre_cti_data_hdu,
                pixel_scales=pixel_scales,
            )
        else:
            raise exc.LayoutException(
                "Cannot estimate pre_cti_data data from non-uniform charge injectiono pattern"
            )

        pre_cti_data = aa.Array1D.no_mask(
            values=pre_cti_data.native, pixel_scales=pixel_scales
        )

        return Dataset1D(
            data=data,
            noise_map=noise_map,
            pre_cti_data=pre_cti_data,
            layout=layout,
            settings_dict=settings_dict,
        )

    def output_to_fits(
        self,
        data_path: Union[Path, str],
        noise_map_path: Optional[Union[Path, str]] = None,
        pre_cti_data_path: Optional[Union[Path, str]] = None,
        overwrite: bool = False,
    ):
        """
        Output the 1D dataset to multiple .fits file.

        For each attribute of the 1D dataset data (e.g. `data`, `noise_map`, `pre_cti_data`) the path to
        the .fits can be specified, with `hdu=0` assumed automatically.

        If the `data` has been masked, the masked data is output to .fits files. A mask can be separately output to
        a file `mask.fits` via the `Mask` objects `output_to_fits` method.

        Parameters
        ----------
        data_path
            The path to the data .fits file where the image data is output (e.g. '/path/to/data.fits').
        noise_map_path
            The path to the noise_map .fits where the noise_map is output (e.g. '/path/to/noise_map.fits').
        pre_cti_data_path
            The path to the pre CTI data .fits file where the pre CTI data is output (e.g. '/path/to/pre_cti_data.fits').
        overwrite
            If `True`, the .fits files are overwritten if they already exist, if `False` they are not and an
            exception is raised.
        """
        self.data.output_to_fits(file_path=data_path, overwrite=overwrite)
        self.noise_map.output_to_fits(file_path=noise_map_path, overwrite=overwrite)
        self.pre_cti_data.output_to_fits(
            file_path=pre_cti_data_path, overwrite=overwrite
        )

    @classmethod
    def from_pixel_line_dict(
        cls,
        pixel_line_dict: dict,
        size: int,
    ) -> "Dataset1D":
        """
        Parse a pixel line output from the warm-pixels script.

        Pixel lines are individual or averaged lines found by searching for
        warm pixels or consistent warm pixels in CCD data. The warm pixel
        and its are extracted and saved as a JSON which can then be loaded
        and fit as part of autocti.

        Parameters
        ----------
        pixel_line_dict
            A dictionary describing a pixel line collection.

            e.g.
            {
                "location": [
                    2,
                    4,
                ],
                "flux": 1234.,
                "data": [
                    5.0,
                    3.0,
                    2.0,
                    1.0,
                ],
                "noise": [
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                ]
            }

            location
                The location of the warm pixel in (row, column) where row is the
                distance to the serial register - 1
            flux
                The computed flux of the warm pixel prior to CTI
            data
                The extracted pixel line. A 1D array where the first entry is
                the warm pixel (FPR) and the remaining entries are the trail
                (EPER)
            noise
                The noise map for the pixel line.
        size
            The size of the CCD. That is, the number of pixels in the parallel
            direction.

        Returns
        -------
            A Dataset1D initialised to represent the pixel line. The pixel line
            and noise are embedded in Array1Ds of the same size as the array in
            the parallel direction
        """
        serial_distance, _ = map(int, pixel_line_dict["location"])

        def make_array(data):
            array = np.zeros(size)
            array[serial_distance : serial_distance + len(data)] = data
            return aa.Array1D.no_mask(array, pixel_scales=0.1)

        return Dataset1D(
            data=make_array(pixel_line_dict["data"]),
            noise_map=make_array(pixel_line_dict["noise"]),
            pre_cti_data=make_array(np.array([pixel_line_dict["flux"]])),
            layout=Layout1D(
                shape_1d=(size,),
                region_list=[
                    aa.Region1D(region=(serial_distance, serial_distance + 1))
                ],
            ),
        )



from autofit import exc
from autoarray.exc import ArrayException, MaskException, RegionException


class DatasetException(Exception):
    pass


class PatternLineException(Exception):
    pass


class ImagingCIException(Exception):
    pass


class ExtractException(Exception):
    pass


class LayoutException(Exception):
    pass


class PlottingException(Exception):
    pass


class FittingException(Exception):
    pass


class PriorException(exc.FitException):
    pass


class ClockerException(Exception):
    pass



import numpy as np
from typing import Optional


class Preloads:
    def __init__(
        self,
        parallel_fast_index_list: Optional[np.ndarray] = None,
        parallel_fast_column_lists: Optional[np.ndarray] = None,
        serial_fast_index_list: Optional[np.ndarray] = None,
        serial_fast_row_lists: Optional[np.ndarray] = None,
        noise_normalization: Optional[float] = None,
    ):
        """
        Class which offers a concise API for settings up the preloads, which before a model-fit are set up via
        inspection of attributes such as the clocker.

        For example, if the clocker's `parallel_fast_mode` is on, this significantly reduces the number of arCTIc
        calls in CTI calibration data. The method inspects every image passed to arctic, extracts
        all unique columns, only passes these to arCTIc and use the output to rebuild the `post_cti_data`. The indexes
        which store unique columns and map them to the post-CTI data can be preloaded in memory to avoid repeated
        calculations in the likelihood function.


        Parameters
        ----------
        parallel_fast_index_list
            The index of a column that is repeated in the pre-cti data. This index corresponds to the first index
            of the repeated columns and this array used to extract the columns from the pre-cti data which are passed
            to arctic.
        parallel_fast_column_lists
            The mapping of every repeated column in `parallel_fast_index_list`  to all other columns which are identical.
             This is used to map the reduced arCTIc output to the post-CTI data.
        serial_fast_index_list
            The index of a row that is repeated in the pre-cti data. This index corresponds to the first index
            of the repeated rows and this array used to extract the rows from the pre-cti data which are passed
            to arctic.
        serial_fast_row_lists
            The mapping of every repeated row in `serial_fast_index_list`  to all other rows which are identical.
            This is used to map the reduced arCTIc output to the post-CTI data.
        noise_normalization
            The noise normalization term of the log likelihood function evaluated in `Analysis` objects. If the
            noise-map is fixed, this can be preloaded as it does not change.

        Returns
        -------
        Preloads
            The preloads object used to skip certain calculations in the log likelihood function.
        """
        self.parallel_fast_index_list = parallel_fast_index_list
        self.parallel_fast_column_lists = parallel_fast_column_lists
        self.serial_fast_index_list = serial_fast_index_list
        self.serial_fast_row_lists = serial_fast_row_lists
        self.noise_normalization = noise_normalization
